from typing import Union, Callable
from numbers import Number

from computegraph import ComputeGraph
from computegraph.types import GraphObject, Function
import jax

from summer2.parameters import Time

from .util import _as_graphobj, piecewise_constant
from . import interpolate


def get_linear_interpolation_function(
    x_pts: jax.Array, y_pts: Union[jax.Array, GraphObject], x_axis=Time
) -> Function:
    """Build a linear interpolator over the x and y points supplied

    Args:
        x_pts: Array of x points over which to interpolate
        y_pts: Values at f(x) for x_pts, over which to interpolate
        x_axis: First argument to this function; defaults to Time.

    Returns:
        Resultant GraphObject Function
    """
    y_pts = _as_graphobj(y_pts)
    yscale = Function(interpolate.get_scale_data, [y_pts])
    curve = interpolate.build_linear_interpolator(x_pts)

    return Function(curve, [x_axis, yscale])


def get_sigmoidal_interpolation_function(
    x_pts: jax.Array, y_pts: Union[jax.Array, GraphObject], x_axis=Time, curvature=16.0
) -> Function:
    """Build a piecewise sigmoidal curve interpolator over the x and y points supplied

    Args:
        x_pts: Array of x points over which to interpolate
        y_pts: Values at f(x) for x_pts, over which to interpolate
        x_axis: First argument to this function; defaults to Time.

        curvature (optional): Curvature of the sigmoid. Defaults to 16.0.

    Returns:
        Function: _description_
    """
    y_pts = _as_graphobj(y_pts)
    yscale = Function(interpolate.get_scale_data, [y_pts])
    curve = interpolate.build_sigmoidal_multicurve(x_pts, curvature=curvature)

    return Function(curve, [x_axis, yscale])


def get_piecewise_scalar_function(
    breakpoints: jax.Array, values: Union[jax.Array, GraphObject], x_axis=Time
) -> Function:
    """Create a Function object to be evaluated in the {x} domain, returning
    a value selected from {values}, the index of which is the x-bounds
    described by {breakpoints}.

    By default x is summer2 Time (time-varying function)

    Args:
        breakpoints: Breakpoints to which x is compared
        values: Array-like of length len(breakpoints) + 1
        x: GraphObject supplying the x value.  Defaults to Time.

    Returns:
        The resulting wrapped Function object
    """
    breakpoints = _as_graphobj(breakpoints)
    values = _as_graphobj(values)
    return Function(piecewise_constant, (x_axis, breakpoints, values))


def _tfunc(f: Function):
    return lambda t, parameters: f(model_variables={"time": t}, parameters=parameters)["out"]


def get_time_callable(f: Function, jit_compile=True) -> Callable:
    """For a Function f which takes Time and Parameters as arguments,
    return a python callable accepting either a number or an an array,
    and a dictionary of parameter values.

    E.g
    returned_func(1.0, parameters={"scale": 0.5})

    Args:
        f: The Time Function
        jit_compile (optional): JIT the function

    Returns:
        The resulting function
    """
    cg = ComputeGraph(f)
    tf = _tfunc(cg.get_callable())
    if jit_compile:
        tf = jax.jit(tf)
    vmapped = jax.vmap(tf, in_axes=(0, None))
    if jit_compile:
        vmapped = jax.jit(vmapped)

    def wrapped(t, parameters):
        if isinstance(t, Number):
            return tf(t, parameters)
        else:
            return vmapped(t, parameters)

    return wrapped



from numbers import Number
from collections.abc import Container
from jax import lax, numpy as jnp
import jax

from computegraph.types import GraphObject, Function, Data
from summer2.parameters import Time


def _as_graphobj(x) -> GraphObject:
    if isinstance(x, GraphObject):
        return x
    elif isinstance(x, Number):
        return x
    else:
        jnp_arr = jnp.array(x)
        return Data(jnp_arr)


def piecewise_function(x, breakpoints, functions):
    index = sum(x >= breakpoints)
    return lax.switch(index, functions, x)


# All
def piecewise_constant(x, breakpoints, values):
    index = sum(x >= breakpoints)
    return values[index]


def windowed_constant(x: float, value: float, window_start: float, window_length: float):
    breakpoints = jnp.array((window_start, window_start + window_length))
    values = jnp.array((0.0, value, 0.0))
    return piecewise_constant(x, breakpoints, values)


def binary_search_ge(x: float, points: jax.Array) -> int:
    """Find the lowest index of the value within points
    to which x is greater than or equal to, using a binary
    search.

    A value of x lower than min(points) will return 0, and higher
    than max(points) will return len(points)-1

    Args:
        x: Value to find
        points: Array to search

    Returns:
        The index value satisying the above conditions
    """

    def cond(state):
        low, high = state
        return (high - low) > 1

    def body(state):
        low, high = state
        midpoint = (0.5 * (low + high)).astype(int)
        update_upper = x < points[midpoint]
        low = jnp.where(update_upper, low, midpoint)
        high = jnp.where(update_upper, midpoint, high)
        return (low, high)

    low, high = lax.while_loop(cond, body, (0, len(points) - 1))
    return lax.cond(x < points[high], lambda: low, lambda: high)

def capture_dict(**kwargs) -> Function:
    """Create a Function that captures a dictionary

    Returns:
        The GraphObject representation of the dict
    """
    return Function(_capture_dict, kwargs=kwargs)

def _capture_dict(**kwargs):
    return kwargs


def capture_array(arr_obj: list, dtype=None) -> Function:
    """Create a GraphObject representation of an (optionally nested) array type,
    declared as a list

    Args:
        arr_obj: A list of the form consumed by (j)np.array(arr_obj)
        dtype (optional): Datatype of the returned array (inferred by default)

    Returns:
        The GraphObject representation of the array
    """
    ishape = _infer_shape(arr_obj)
    
    if len(ishape) > 1:
        # arr_obj is a nested listed, which we must flatten to expose args to the graph
        flattened = _flatten_arr_obj(arr_obj)
        
        def _capture_array_reshape(*args):
            return jnp.array(args, dtype=dtype).reshape(ishape)
        
        return Function(_capture_array_reshape, args=flattened)
    else:
        # Simple instance of 1d array
        def _capture_array(*args):
            return jnp.array(args, dtype=dtype)
    
        return Function(_capture_array, args=arr_obj)

def _infer_shape(arr_obj: list, cur_shape=()):
    new_shape = (*cur_shape, len(arr_obj))
    if isinstance(arr_obj[0], list):
        msg = "All subarrays must be lists"
        assert all([isinstance(sub_arr, list) for sub_arr in arr_obj]), msg
        return _infer_shape(arr_obj[0], new_shape)
    return new_shape

def _flatten_arr_obj(arr_obj: list, cur_obj=None):
    if cur_obj is None:
        cur_obj = []
    for v in arr_obj:
        if isinstance(v, list):
            _flatten_arr_obj(v, cur_obj)
        else:
            cur_obj.append(v)
    return cur_obj


from summer2 import CompartmentalModel
from summer2.parameters import Parameter as param
from summer2.parameters import DerivedOutput

def sir():
    m = CompartmentalModel([0,100], ["S","I","R"], "I")
    m.set_initial_population({"S": 990.0, "I": 10.0})
    m.add_infection_frequency_flow("infection", param("contact_rate"), "S", "I")
    m.add_transition_flow("recovery", param("recovery_rate"), "I", "R")
    m.request_output_for_flow("incidence", "infection")
    m.set_default_parameters({"contact_rate": 0.4, "recovery_rate": 0.1})
    return m

